# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <phil@canary.md>
# date: 2015/04/02
# copy: (C) Copyright 2015-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import morph

from .decoder import SortValidator, SmartSort

#------------------------------------------------------------------------------
class Mapper(object):
  '''
  A `Mapper` selects the result set to be paginated from the
  response and transforms the response to contain the paginated result
  set and the pagination meta-information.
  '''

  #----------------------------------------------------------------------------
  def __init__(self, target=None, *args, **kw):
    super(Mapper, self).__init__(*args, **kw)
    self.target = None

  #----------------------------------------------------------------------------
  def extend(self, *args, **kw):
    params = dict(target=self.target)
    for arg in args:
      params.update(arg)
    params.update(kw)
    return self.__class__(**params)

  #----------------------------------------------------------------------------
  def get(self, p8n, result):
    return self.resolve(p8n, result)()

  #----------------------------------------------------------------------------
  def resolve(self, p8n, result):
    '''
    Returns a function that is expected to be called in one of the
    following ways:

    * as a "getter" (with no arguments)
    * as a "setter" (with exactly one argument)
    * returns the target path (with exactly two arguments)
    '''
    # todo: this `exactly two arguments` is ridiculous... change!
    if self.target is not None:
      return self.resolve_target(p8n, result)
    if not morph.isdict(result):
      def _resolve(*args):
        if len(args) <= 0:
          return result
        if len(args) == 1:
          return dict(((p8n.paginator.result_name, args[0]),))
        if len(args) == 2:
          return p8n.paginator.result_name
      return _resolve
    if len(result.keys()) != 1:
      raise ValueError(
        'Pagination of multi-key dictionaries requires setting the'
        ' pagination mapper "target" attribute')
    key = result.keys()[0]
    def _resolve(*args):
      if len(args) <= 0:
        return result[key]
      if len(args) == 1:
        return dict([(key, args[0])])
      if len(args) == 2:
        return key
    return _resolve

  #----------------------------------------------------------------------------
  def resolve_target(self, p8n, result):
    # todo: support list-index style notation as well, eg:
    #       ``foo-1.bar`` would resolve to the ``"here"`` element in:
    #       {foo: [{bar: 'no'}, {bar: 'here'}, {bar: 'nada'}]}
    # TODO: do error checking for missing keys...
    container = result
    keys = self.target.split('.')
    for key in keys[:-1]:
      container = container.get(key)
    key = keys[-1]
    def _resolve(*args):
      if len(args) <= 0:
        return container[key]
      if len(args) == 1:
        container[key] = args[0]
        return result
      if len(args) == 2:
        return self.target
    return _resolve

  #----------------------------------------------------------------------------
  def put(self, p8n, result, value):
    result = self.put_data(p8n, result, value)
    result = self.put_meta(p8n, result, value)
    return result

  #----------------------------------------------------------------------------
  def put_data(self, p8n, result, value):
    resolver = self.resolve(p8n, result)
    value[1]['attribute'] = resolver(None, None)
    return resolver(value[0])

  #----------------------------------------------------------------------------
  def put_meta(self, p8n, result, value):
    page = dict()
    if 'count' in value[1]:
      page[p8n.paginator.count_name]  = value[1]['count']
    page[p8n.paginator.offset_name] = p8n.offset
    page[p8n.paginator.limit_name]  = p8n.limit
    sort = SortValidator.encode(p8n.sort)
    if sort != ( SmartSort.MARK
                 if p8n.paginator.sort_default is SmartSort
                 else p8n.paginator.sort_default ):
      page[p8n.paginator.sort_name] = sort
    page[p8n.paginator.attribute_name] = value[1].get(
      'attribute', p8n.paginator.result_name) or p8n.paginator.result_name
    try:
      ret = dict(result)
    except ValueError:
      # todo: what the ... ?
      ret = dict(result=result)
    ret[p8n.paginator.page_name] = page
    return ret


#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <phil@canary.md>
# date: 2015/04/02
# copy: (C) Copyright 2015-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

from collections import OrderedDict

import six
import morph
import sqlalchemy
import sqlalchemy.orm

from .decoder import SmartSort

#------------------------------------------------------------------------------
class Engine(object):
  '''
  The `Engine` performs the actual pagination work of narrowing down
  the result set to the desired window.
  '''

  #----------------------------------------------------------------------------
  def __init__(self, comparers={}, *args, **kw):
    super(Engine, self).__init__(*args, **kw)
    self.comparers = OrderedDict()
    if morph.isstr(comparers):
      comparers = [comparers]
    if morph.isseq(comparers) and len(comparers) > 0 \
        and not morph.isseq(comparers[0]):
      self.comparers.update((k, k) for k in comparers)
    else:
      self.comparers.update(comparers)

  #----------------------------------------------------------------------------
  def extend(self, *args, **kw):
    params = dict(comparers=self.comparers)
    for arg in args:
      params.update(arg)
    params.update(kw)
    return self.__class__(**params)

  #----------------------------------------------------------------------------
  def apply(self, p8n, result):
    '''
    Returns a two-element tuple of ``(narrowed_value, page_attributes)``.
    '''
    if isinstance(result, sqlalchemy.orm.Query):
      return self.apply_sqlalchemy_orm_query_query(p8n, result)
    if isinstance(result, (list, tuple)):
      return self.apply_list(p8n, result)
    try: result = tuple(result)
    except: pass
    try:
      implementation = getattr(self, 'apply_' + self.route(result))
    except:
      raise ValueError('No pagination available for %r' % (type(result),))
    return implementation(p8n, result)

  #----------------------------------------------------------------------------
  def route(self, value):
    ret = type(value)
    if ret.__module__ in ('__builtin__', 'builtin'):
      return ret.__name__.lower()
    return '.'.join([ret.__module__, ret.__name__]).lower()

  #----------------------------------------------------------------------------
  def sorters(self, p8n):
    if p8n.sort is not SmartSort:
      return p8n.sort
    return [(key, True) for key in self.comparers.keys()]

  #----------------------------------------------------------------------------
  def apply_list(self, p8n, value):
    sorters = self.sorters(p8n)
    def sortfunc(a, b):
      for meth, asc in sorters:
        spec = self.comparers[meth]
        if spec is None:
          continue
        elif six.callable(spec):
          try:
            ret = spec(p8n, value, a, b)
          except TypeError:
            ret = spec(a, b)
        elif morph.isstr(spec):
          try:
            ret = cmp(getattr(a, spec), getattr(b, spec))
          except AttributeError:
            ret = cmp(a[spec], b[spec])
        else:
          raise ValueError(
            'pagination sort comparer must be a callable,'
            ' an attribute, or an item key')
        if ret == 0:
          continue
        if asc:
          return ret
        if ret < 0:
          return 1
        return -1
      return 0
    value = sorted(value, cmp=sortfunc)
    count = len(value)
    if p8n.limit > 0:
      value = value[p8n.offset : p8n.offset + p8n.limit]
    else:
      value = value[p8n.offset : ]
    return (value, dict(count=count))

  #----------------------------------------------------------------------------
  def apply_sqlalchemy_orm_query_query(self, p8n, query):
    for meth, asc in self.sorters(p8n):
      spec = self.comparers[meth]
      if spec is None:
        continue
      elif six.callable(spec):
        query = spec(pagination=p8n, query=query, method=meth, ascending=asc)
      elif morph.isstr(spec):
        query = query.order_by(spec + ( '' if asc else ' DESC' ))
      else:
        query = query.order_by(spec if asc else sqlalchemy.desc(spec))
    count = query.count()
    query = query.offset(p8n.offset)
    if p8n.limit > 0:
      query = query.limit(p8n.limit)
    return (query, dict(count=count))

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <pjg.github@ubergrabner.net>
# date: 2015/03/28
# copy: (C) Copyright 2015-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import types
import unittest

from aadict import aadict
import morph
from pyramid.request import Request
from six.moves.urllib.parse import urlencode

#------------------------------------------------------------------------------
class TestListPagination(unittest.TestCase):

  maxDiff = None

  #----------------------------------------------------------------------------
  @staticmethod
  def request(url='/', **kw):
    # TODO: *much* better would be to override request.params... but it
    #       won't allow that! ugh. if that is fixed, go through and removed
    #       most of:
    #       - decoder={'request_param': 'data'}
    #       - param='data'
    #       - `peepsd` and `n10d`
    data = None
    if kw:
      for k, v in kw.items():
        if not morph.isstr(v) or not morph.isstr(k):
          data = kw
          break
      if data is None:
        if '?' not in url:
          url += '?'
        url += urlencode(kw)
    ret = Request.blank(url)
    if data is not None:
      ret.data = data
    return ret

  #----------------------------------------------------------------------------
  def test_invalid_sort_value(self):
    from .paginator import paginate
    import formencode.api
    @paginate(decoder={'request_param': 'data'})
    def n30(request):
      return list(range(30))
    with self.assertRaises(formencode.api.Invalid) as cm:
      n30(self.request(**{'page.sort': 88}))
    self.assertEqual(
      str(cm.exception),
      'page.sort: Please specify a string or list of strings')

  #----------------------------------------------------------------------------
  def test_invalid_sort_method(self):
    from .paginator import paginate
    import formencode.api
    @paginate
    def n30(request):
      return list(range(30))
    with self.assertRaises(formencode.api.Invalid) as cm:
      n30(self.request(**{'page.sort': 'no-such-method'}))
    self.assertEqual(
      str(cm.exception),
      'page.sort: Invalid sorting method "no-such-method"')

  #----------------------------------------------------------------------------
  def test_invalid_sort_method_structured(self):
    from .paginator import paginate
    import formencode.api
    @paginate(decoder={'structured': True, 'request_param': 'data'})
    def n30(request):
      return list(range(30))
    with self.assertRaises(formencode.api.Invalid) as cm:
      n30(self.request(page=dict(sort='no-such-method')))
    self.assertEqual(
      str(cm.exception),
      'page: sort: Invalid sorting method "no-such-method"')

  #----------------------------------------------------------------------------
  def test_decoder_structured(self):
    from .paginator import paginate
    @paginate(decoder={'structured': True, 'request_param': 'data'})
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request(page=dict(limit=10))),
      dict(
        result = list(range(10)),
        page   = dict(offset=0, limit=10, count=30, attribute='result')))

  #----------------------------------------------------------------------------
  def test_list_default(self):
    from .paginator import paginate
    @paginate
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request()),
      dict(
        result = list(range(25)),
        page   = dict(offset=0, limit=25, count=30, attribute='result')))

  #----------------------------------------------------------------------------
  def test_list_default_structured(self):
    from .paginator import paginate
    @paginate(decoder={'structured': True})
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request()),
      dict(
        result = list(range(25)),
        page   = dict(offset=0, limit=25, count=30, attribute='result')))

  #----------------------------------------------------------------------------
  def test_list_offset(self):
    from .paginator import paginate
    @paginate
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request(**{'page.offset': '10'})),
      dict(
        result = list(range(10, 30)),
        page   = dict(offset=10, limit=25, count=30, attribute='result')))

  #----------------------------------------------------------------------------
  def test_list_offsetlimit(self):
    from .paginator import paginate
    @paginate
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request(**{'page.offset': '10', 'page.limit': '5'})),
      dict(
        result = list(range(10, 15)),
        page   = dict(offset=10, limit=5, count=30, attribute='result')))

  #----------------------------------------------------------------------------
  def test_list_limit_default(self):
    from .paginator import paginate
    @paginate(limit_default=20)
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request()),
      dict(
        result = list(range(20)),
        page   = dict(offset=0, limit=20, count=30, attribute='result')))

  #----------------------------------------------------------------------------
  def test_list_limit_none(self):
    from .paginator import paginate
    @paginate(limit_default=20, decoder={'request_param': 'data'})
    def n300(request):
      return list(range(300))
    self.assertEqual(
      n300(self.request(**{'page.limit': 0})),
      dict(
        result = list(range(300)),
        page   = dict(offset=0, limit=0, count=300, attribute='result')))

  #----------------------------------------------------------------------------
  def test_list_limit_param(self):
    from .paginator import paginate
    @paginate(limit_default=20, decoder={'request_param': 'data'})
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request(**{'page.limit': 10})),
      dict(
        result = list(range(10)),
        page   = dict(offset=0, limit=10, count=30, attribute='result')))

  #----------------------------------------------------------------------------
  def test_list_sort_comparers(self):
    from .paginator import paginate
    def sort_by_evenodd(p8n, result, a, b):
      if ( a % 2 ) == ( b % 2 ):
        return 0
      if a % 2 == 0:
        return -1
      return 1
    @paginate(comparers={'evenodd': sort_by_evenodd, 'num': cmp})
    def n10(request):
      return list(range(10))
    @paginate(comparers={'evenodd': sort_by_evenodd, 'num': cmp}, decoder={'request_param': 'data'})
    def n10d(request):
      return list(range(10))
    self.assertEqual(
      n10(self.request(**{'page.sort': 'evenodd'})),
      dict(
        result = [0, 2, 4, 6, 8, 1, 3, 5, 7, 9],
        page   = dict(offset=0, limit=25, count=10, sort='evenodd', attribute='result')))
    self.assertEqual(
      n10d(self.request(**{'page.sort': 'evenodd', 'page.limit': 6})),
      dict(
        result = [0, 2, 4, 6, 8, 1],
        page   = dict(offset=0, limit=6, count=10, sort='evenodd', attribute='result')))
    self.assertEqual(
      n10(self.request(**{'page.sort': 'evenodd-'})),
      dict(
        result = [1, 3, 5, 7, 9, 0, 2, 4, 6, 8],
        page   = dict(offset=0, limit=25, count=10, sort='evenodd-', attribute='result')))
    self.assertEqual(
      n10(self.request(**{'page.sort': 'evenodd-,num'})),
      dict(
        result = [1, 3, 5, 7, 9, 0, 2, 4, 6, 8],
        page   = dict(offset=0, limit=25, count=10, sort='evenodd-,num', attribute='result')))
    self.assertEqual(
      n10(self.request(**{'page.sort': 'evenodd-,num-'})),
      dict(
        result = [9, 7, 5, 3, 1, 8, 6, 4, 2, 0],
        page   = dict(offset=0, limit=25, count=10, sort='evenodd-,num-', attribute='result')))

  #----------------------------------------------------------------------------
  def test_list_sort_default(self):
    from .paginator import paginate
    def sort_by_evenodd(p8n, result, a, b):
      if ( a % 2 ) == ( b % 2 ):
        return 0
      if a % 2 == 0:
        return -1
      return 1
    @paginate(comparers=(('evenodd', sort_by_evenodd), ('num', cmp)))
    def n10(request):
      return [3, 7, 1, 8, 0, 4, 6, 9, 2, 5]
    self.assertEqual(
      n10(self.request()),
      dict(
        result = [0, 2, 4, 6, 8, 1, 3, 5, 7, 9],
        page   = dict(offset=0, limit=25, count=10, attribute='result')))
    @paginate(comparers=(('num', cmp), ('evenodd', sort_by_evenodd)))
    def n10(request):
      return [3, 7, 1, 8, 0, 4, 6, 9, 2, 5]
    self.assertEqual(
      n10(self.request()),
      dict(
        result = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        page   = dict(offset=0, limit=25, count=10, attribute='result')))
    @paginate(sort_default='', comparers=(('num', cmp), ('evenodd', sort_by_evenodd)))
    def n10r(request):
      return [3, 7, 1, 8, 0, 4, 6, 9, 2, 5]
    self.assertEqual(
      n10r(self.request()),
      dict(
        result = [3, 7, 1, 8, 0, 4, 6, 9, 2, 5],
        page   = dict(offset=0, limit=25, count=10, attribute='result')))
    @paginate(comparers=(('num', cmp), ('evenodd', sort_by_evenodd)))
    def n10(request):
      return [3, 7, 1, 8, 0, 4, 6, 9, 2, 5]
    self.assertEqual(
      n10(self.request(**{'page.sort': ''})),
      dict(
        result = [3, 7, 1, 8, 0, 4, 6, 9, 2, 5],
        page   = dict(offset=0, limit=25, count=10, sort='', attribute='result')))

  #----------------------------------------------------------------------------
  def test_list_sort_override_default(self):
    from .paginator import paginate
    def sort_by_evenodd(p8n, result, a, b):
      if ( a % 2 ) == ( b % 2 ):
        return 0
      if a % 2 == 0:
        return -1
      return 1
    @paginate(sort_default='evenodd-,num-', comparers={'evenodd': sort_by_evenodd, 'num': cmp})
    def n10(request):
      return list(range(10))
    self.assertEqual(
      n10(self.request()),
      dict(
        result = [9, 7, 5, 3, 1, 8, 6, 4, 2, 0],
        page   = dict(offset=0, limit=25, count=10, attribute='result')))

  #----------------------------------------------------------------------------
  def test_paginate_extend(self):
    from .paginator import paginate
    paginate = paginate(limit_default=20)
    @paginate
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request()),
      dict(
        result = list(range(20)),
        page   = dict(offset=0, limit=20, count=30, attribute='result')))
    paginate = paginate(limit_default=15)
    @paginate
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request()),
      dict(
        result = list(range(15)),
        page   = dict(offset=0, limit=15, count=30, attribute='result')))
    @paginate(limit_default=10)
    def n30_10(request):
      return list(range(30))
    self.assertEqual(
      n30_10(self.request()),
      dict(
        result = list(range(10)),
        page   = dict(offset=0, limit=10, count=30, attribute='result')))
    self.assertEqual(
      n30(self.request()),
      dict(
        result = list(range(15)),
        page   = dict(offset=0, limit=15, count=30, attribute='result')))

  #----------------------------------------------------------------------------
  def test_sort_attribute_name_map(self):
    from .paginator import paginate
    paginate = paginate(sort_default='name,age-', comparers={'name': 'name', 'age': 'age'})
    @paginate
    def peeps(request):
      return [
        aadict(id=1, name='zeta', age=8),
        aadict(id=2, name='delt', age=2),
        aadict(id=3, name='zeta', age=4),
        aadict(id=4, name='acrn', age=6),
      ]
    self.assertEqual(
      peeps(self.request()),
      dict(
        result = [
          aadict(id=4, name='acrn', age=6),
          aadict(id=2, name='delt', age=2),
          aadict(id=1, name='zeta', age=8),
          aadict(id=3, name='zeta', age=4),
        ],
        page   = dict(offset=0, limit=25, count=4, attribute='result')))
    @paginate(sort_default='name,age')
    def peeps(request):
      return [
        aadict(id=1, name='zeta', age=8),
        aadict(id=2, name='delt', age=2),
        aadict(id=3, name='zeta', age=4),
        aadict(id=4, name='acrn', age=6),
      ]
    self.assertEqual(
      peeps(self.request()),
      dict(
        result = [
          aadict(id=4, name='acrn', age=6),
          aadict(id=2, name='delt', age=2),
          aadict(id=3, name='zeta', age=4),
          aadict(id=1, name='zeta', age=8),
        ],
        page   = dict(offset=0, limit=25, count=4, attribute='result')))

  #----------------------------------------------------------------------------
  def test_sort_attribute_name_list(self):
    from .paginator import paginate
    paginate = paginate(sort_default='name,age-', comparers=['name', 'age'])
    @paginate
    def peeps(request):
      return [
        aadict(id=1, name='zeta', age=8),
        aadict(id=2, name='delt', age=2),
        aadict(id=3, name='zeta', age=4),
        aadict(id=4, name='acrn', age=6),
      ]
    self.assertEqual(
      peeps(self.request()),
      dict(
        result = [
          aadict(id=4, name='acrn', age=6),
          aadict(id=2, name='delt', age=2),
          aadict(id=1, name='zeta', age=8),
          aadict(id=3, name='zeta', age=4),
        ],
        page   = dict(offset=0, limit=25, count=4, attribute='result')))

  #----------------------------------------------------------------------------
  def test_map_item(self):
    from .paginator import paginate
    @paginate(
      limit_default=3,
      force_list=False,
      map_item=lambda item, **kw: dict(i=item, s=str(item)))
    def n30(request):
      return list(range(30))
    ret = n30(self.request())
    self.assertIsInstance(ret['result'], types.GeneratorType)
    ret['result'] = list(ret['result'])
    self.assertEqual(
      ret,
      dict(
        result = [dict(i=0, s='0'), dict(i=1, s='1'), dict(i=2, s='2')],
        page   = {'count': 30, 'attribute': 'result', 'limit': 3, 'offset': 0}))

  #----------------------------------------------------------------------------
  def test_map_item_backwardcompatibility(self):
    from .paginator import paginate
    @paginate(
      limit_default=3,
      map_item=lambda item, **kw: dict(i=item, s=str(item)))
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request()),
      dict(
        result = (dict(i=0, s='0'), dict(i=1, s='1'), dict(i=2, s='2')),
        page   = {'count': 30, 'attribute': 'result', 'limit': 3, 'offset': 0}))

  #----------------------------------------------------------------------------
  def test_map_list(self):
    from .paginator import paginate
    @paginate(
      limit_default=5,
      map_list=lambda value, attributes, **kw: (value[:-1], attributes))
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request()),
      dict(
        result = [0, 1, 2, 3],
        page   = {'count': 30, 'attribute': 'result', 'limit': 5, 'offset': 0}))

  #----------------------------------------------------------------------------
  def test_map_return(self):
    from .paginator import paginate
    @paginate(
      limit_default=5,
      map_return=lambda value, **kw: dict(wrapper=value))
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request()),
      dict(wrapper=dict(
        result = [0, 1, 2, 3, 4],
        page   = {'count': 30, 'attribute': 'result', 'limit': 5, 'offset': 0})))

  #----------------------------------------------------------------------------
  def test_map_return_receives_pagination_state(self):
    from .paginator import paginate, Paginator
    @paginate(
      limit_default=5,
      map_return=lambda value, state, **kw: state)
    def n30(request):
      return list(range(30))
    state = n30(self.request())
    self.assertIsInstance(state, aadict)
    self.assertIsInstance(state.paginator, Paginator)

  #----------------------------------------------------------------------------
  def test_keep_items_default(self):
    from .paginator import paginate
    @paginate(
      limit_default=5,
      map_return=lambda value, state, **kw: dict(wrapper=value, items=state.get('items')))
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request()),
      dict(items=None, wrapper=dict(
        result = [0, 1, 2, 3, 4],
        page   = {'count': 30, 'attribute': 'result', 'limit': 5, 'offset': 0})))

  #----------------------------------------------------------------------------
  def test_keep_items_enabled(self):
    from .paginator import paginate
    @paginate(
      limit_default=5,
      keep_items=True,
      map_return=lambda value, state, **kw: dict(wrapper=value, items=state.get('items')))
    def n30(request):
      return list(range(30))
    self.assertEqual(
      n30(self.request()),
      dict(items=(0, 1, 2, 3, 4), wrapper=dict(
        result = (0, 1, 2, 3, 4),
        page   = {'count': 30, 'attribute': 'result', 'limit': 5, 'offset': 0})))

  #----------------------------------------------------------------------------
  def test_tuple(self):
    from .paginator import paginate
    @paginate(limit_default=3)
    def t30(request):
      return tuple(range(30))
    self.assertEqual(
      t30(self.request()),
      dict(
        result = [0, 1, 2],
        page   = {'count': 30, 'attribute': 'result', 'limit': 3, 'offset': 0}))


#------------------------------------------------------------------------------
class TestSqlalchemyPagination(unittest.TestCase):

  maxDiff = None

  #----------------------------------------------------------------------------
  @staticmethod
  def request(*args, **kw):
    return TestListPagination.request(*args, **kw)

  #----------------------------------------------------------------------------
  def makedb(self):
    import sqlalchemy as sa
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy.orm import sessionmaker
    engine = sa.create_engine('sqlite://')
    Base = declarative_base()
    class Person(Base):
      __tablename__ = 'persons'
      id = sa.Column(sa.Integer, primary_key=True)
      name = sa.Column(sa.String)
      age = sa.Column(sa.Integer)
    class User(Base):
      __tablename__ = 'users'
      id = sa.Column(sa.Integer, primary_key=True)
      person_id = sa.Column(sa.Integer)
    Base.metadata.create_all(engine) 
    Session = sessionmaker()
    Session.configure(bind=engine)
    session = Session()
    model = aadict(engine=engine, Person=Person, User=User, session=session)
    return model

  #----------------------------------------------------------------------------
  def populate(self, model):
    model.session.add(model.Person(id=1, name='zeta', age=8))
    model.session.add(model.Person(id=2, name='delt', age=2))
    model.session.add(model.Person(id=3, name='zeta', age=4))
    model.session.add(model.Person(id=4, name='acrn', age=6))
    model.session.add(model.User(id=4, person_id=1))
    model.session.add(model.User(id=5, person_id=2))
    model.session.add(model.User(id=3, person_id=3))
    model.session.add(model.User(id=6, person_id=4))
    model.session.commit()
    return model

  #----------------------------------------------------------------------------
  def query(self, model, request, param='params'):
    ret = model.session.query(model.Person)
    if 'minage' in getattr(request, param):
      ret = ret.filter(
        model.Person.age >= int(getattr(request, param).get('minage')))
    return ret

  #----------------------------------------------------------------------------
  def dictify(self, result, pick=('id', 'name', 'age'), pluck=None):
    if pick:
      result['result'] = [morph.pick(peep, *pick) for peep in result['result']]
    if pluck:
      result['result'] = [peep[pluck] for peep in result['result']]
    return result

  #----------------------------------------------------------------------------
  def test_sqlalchemy(self):
    model = self.populate(self.makedb())
    from .paginator import paginate
    @paginate
    def peeps(request):
      return self.query(model, request)
    @paginate(decoder={'request_param': 'data'})
    def peepsd(request):
      return self.query(model, request, param='data')
    self.assertEqual(
      self.dictify(peeps(self.request()), pluck='name'),
      dict(
        page   = dict(offset=0, limit=25, count=4, attribute='result'),
        result = ['zeta', 'delt', 'zeta', 'acrn']))
    self.assertEqual(
      self.dictify(peepsd(self.request(minage=6)), pluck='id'),
      dict(
        page   = dict(offset=0, limit=25, count=2, attribute='result'),
        result = [1, 4]))

  #----------------------------------------------------------------------------
  def test_sqlalchemy_limit(self):
    model = self.populate(self.makedb())
    from .paginator import paginate
    @paginate(decoder={'request_param': 'data'})
    def peeps(request):
      return self.query(model, request)
    self.assertEqual(
      self.dictify(peeps(self.request(**{'page.limit': 2})), pluck='id'),
      dict(
        page   = dict(offset=0, limit=2, count=4, attribute='result'),
        result = [1, 2]))

  #----------------------------------------------------------------------------
  def test_sqlalchemy_offset(self):
    model = self.populate(self.makedb())
    from .paginator import paginate
    @paginate(decoder={'request_param': 'data'})
    def peeps(request):
      return self.query(model, request)
    self.assertEqual(
      self.dictify(peeps(self.request(**{'page.offset': 1})), pluck='id'),
      dict(
        page   = dict(offset=1, limit=25, count=4, attribute='result'),
        result = [2, 3, 4]))

  #----------------------------------------------------------------------------
  def test_sqlalchemy_offsetlimit(self):
    model = self.populate(self.makedb())
    from .paginator import paginate
    @paginate(decoder={'request_param': 'data'})
    def peeps(request):
      return self.query(model, request)
    self.assertEqual(
      self.dictify(peeps(self.request(**{'page.offset': 1, 'page.limit': 1})), pluck='id'),
      dict(
        page   = dict(offset=1, limit=1, count=4, attribute='result'),
        result = [2]))

  #----------------------------------------------------------------------------
  def test_sqlalchemy_sort(self):
    model = self.populate(self.makedb())
    from .paginator import paginate
    @paginate(sort_default='name', comparers={'name': 'name', 'age': 'age'})
    def peeps(request):
      return self.query(model, request)
    self.assertEqual(
      self.dictify(peeps(self.request()), pluck='name'),
      dict(
        page   = dict(offset=0, limit=25, count=4, attribute='result'),
        result = ['acrn', 'delt', 'zeta', 'zeta']))
    self.assertEqual(
      self.dictify(peeps(self.request(**{'page.sort': 'name-'})), pluck='name'),
      dict(
        page   = dict(offset=0, limit=25, count=4, sort='name-', attribute='result'),
        result = ['zeta', 'zeta', 'delt', 'acrn']))
    self.assertEqual(
      self.dictify(peeps(self.request(**{'page.sort': 'age'})), pluck='id'),
      dict(
        page   = dict(offset=0, limit=25, count=4, sort='age', attribute='result'),
        result = [2, 3, 4, 1]))
    self.assertEqual(
      self.dictify(peeps(self.request(**{'page.sort': 'age-'})), pluck='id'),
      dict(
        page   = dict(offset=0, limit=25, count=4, sort='age-', attribute='result'),
        result = [1, 4, 3, 2]))
    self.assertEqual(
      self.dictify(peeps(self.request(**{'page.sort': 'name,age'})), pluck='id'),
      dict(
        page   = dict(offset=0, limit=25, count=4, sort='name,age', attribute='result'),
        result = [4, 2, 3, 1]))
    self.assertEqual(
      self.dictify(peeps(self.request(**{'page.sort': 'name,age-'})), pluck='id'),
      dict(
        page   = dict(offset=0, limit=25, count=4, sort='name,age-', attribute='result'),
        result = [4, 2, 1, 3]))

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <pjg.github@ubergrabner.net>
# date: 2015/03/20
# copy: (C) Copyright 2015-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import six
from pyramid.request import Request
from aadict import aadict

from .decoder import Decoder, SmartSort
from .mapper import Mapper
from .engine import Engine

#------------------------------------------------------------------------------
def _extend(klass, base, spec):
  if spec is None:
    if base:
      return base
    return klass()
  if isinstance(spec, klass):
    return spec
  if base:
    return base.extend(spec)
  return klass().extend(spec)


#------------------------------------------------------------------------------
class Paginator(object):

  DEFAULTS = dict(
    page_name        = 'page',          # pagination parameters namespace
    offset_name      = 'offset',        # `offset` parameter name
    offset_default   = 0,               # `offset` default value
    limit_name       = 'limit',         # `limit` parameter name
    limit_default    = 25,              # `limit` default value
    sort_name        = 'sort',          # `sort` parameter name
    sort_default     = SmartSort,       # `sort` default value
    count_name       = 'count',         # `count` response parameter name
    attribute_name   = 'attribute',     # `attribute` response parameter name
    result_name      = 'result',        # `result` response namespace
    request_name     = 'pagination',    # pyramid request attribute name for pagination
    keep_items       = False,           # keep the narrowed set in `state.items`?
    force_list       = True,            # force the result set to be a list/tuple?
    map_item         = None,            # per-item result callback hook
    map_list         = None,            # entire result callback hook
    map_return       = None,            # return value callback hook
  )

  #----------------------------------------------------------------------------
  def __init__(self,
               decoder   = None,  # extract pagination parameters
               mapper    = None,  # what to paginate
               engine    = None,  # apply sorting & paging
               *args, **kw):
    for key, defval in Paginator.DEFAULTS.items():
      if key in kw:
        setattr(self, key, kw.pop(key))
      else:
        setattr(self, key, defval)
    self.decoder  = _extend(Decoder, None, decoder)
    self.mapper   = _extend(Mapper,  None, mapper)
    self.engine   = _extend(Engine,  None, engine)
    self.schema   = None  # used by :class:`Decoder`
    if 'comparers' in kw:
      self.engine   = self.engine.extend(comparers=kw.pop('comparers'))
    super(Paginator, self).__init__(*args, **kw)

  #----------------------------------------------------------------------------
  def extend(self, **kw):
    kw['decoder'] = _extend(Decoder, self.decoder, kw.pop('decoder', None))
    kw['mapper']  = _extend(Mapper,  self.mapper,  kw.pop('mapper',  None))
    kw['engine']  = _extend(Engine,  self.engine,  kw.pop('engine',  None))
    for key, defval in Paginator.DEFAULTS.items():
      if key not in kw:
        if getattr(self, key, None) != defval:
          kw[key] = getattr(self, key, None)
    return self.__class__(**kw)

  #----------------------------------------------------------------------------
  def __call__(self, *args, **kw):
    if len(args) != 1 or kw or not six.callable(args[0]):
      # `@paginate` was called with arguments -- force-standardize
      # todo: the problem with this implementation is that it cannot
      #       detect if you passed a function as an argument, eg.:
      #         @paginate(lambda x: foo)
      #       true, Paginate.extend() does not accept that...
      #       *and* it's not very intuitive what would be done with it...
      #       but *still* !
      #       (poor design of the python decorator API, IMHO)
      return self.extend(*args, **kw)
    func = args[0]
    def _wrapped(*args, **kw):
      return self.paginate(func, *args, **kw)
    _wrapped.__doc__ = func.__doc__
    return _wrapped

  #----------------------------------------------------------------------------
  def paginate(self, handler, *args, **kw):
    # todo: i don't particulary like this searching for the "Request"...
    request = [arg for arg in args if isinstance(arg, Request)][0]
    p8n = aadict(paginator=self, request=request)
    p8n.update(self.decoder.decode(p8n))
    setattr(request, self.request_name, p8n)
    result = handler(*args, **kw)
    value  = self.mapper.get(p8n, result)
    value  = self.engine.apply(p8n, value)
    if self.keep_items:
      p8n['items'] = tuple(value[0])
      value = ( p8n['items'], value[1] )
    if self.map_item:
      def mapped_item_generator(items, attrs):
        for item in items:
          yield self.map_item(
            state=p8n, result=result, value=items, item=item, attributes=attrs)
      value = ( mapped_item_generator(value[0], value[1]), value[1] )
    if self.map_list:
      value = self.map_list(
        state=p8n, result=result, value=value[0], attributes=value[1])
    if self.force_list and not isinstance(value[0], (tuple, list)):
      value = ( tuple(value[0]), value[1] )
    value = self.mapper.put(p8n, result, value)
    if self.map_return:
      return self.map_return(state=p8n, result=result, value=value)
    return value

  #----------------------------------------------------------------------------
  @staticmethod
  def get_paginator(handle):
    if not callable(handle):
      return None
    if isinstance(handle, Paginator):
      return handle
    try:
      handle = six.get_function_closure(handle)
    except AttributeError:
      handle = None
    for cell in reversed(handle or ()):
      cell = Paginator.get_paginator(cell.cell_contents)
      if cell is not None:
        return cell
    return None


#------------------------------------------------------------------------------
paginate = Paginator()

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: Philip J Grabner <phil@canary.md>
# date: 2015/04/02
# copy: (C) Copyright 2015-EOT Canary Health, Inc., All Rights Reserved.
#------------------------------------------------------------------------------

import formencode
from formencode import validators
import morph

#------------------------------------------------------------------------------
class _SmartSort(object):
  MARK = '*'
  def __iter__(self):
    yield (SmartSort, True)
SmartSort = _SmartSort()

#------------------------------------------------------------------------------
class RenameFields(formencode.validators.FormValidator):
  def __init__(self, name_map, *args, **kw):
    formencode.validators.FormValidator.__init__(
      self, name_map=name_map, *args, **kw)
  def _to_python(self, value, state):
    return self.keymap(value, self.name_map)
  def _from_python(self, value, state):
    return self.keymap(value, {v: k for k, v in self.nmap})
  def keymap(self, value, nmap):
    return {nmap[k] if k in nmap else k: v for k, v in value.items()}


#------------------------------------------------------------------------------
class SortValidator(formencode.validators.FancyValidator):
  messages = {
    'bad_type'   : 'Please specify a string or list of strings',
    'bad_method' : 'Invalid sorting method "%(method)s"',
    'bad_decode' : ( 'sort specification must be a string,'
                     ' list of strings, '
                     ' or list of (string, bool) tuples' ),
  }
  def _to_python(self, value, state):
    try:
      return self.decode(value)
    except:
      raise formencode.api.Invalid(
        self.message('bad_type', state), value, state)
  @staticmethod
  def decode(spec):
    if spec is None:
      return ()
    if spec is SmartSort:
      return SmartSort
    if morph.isstr(spec):
      if spec.strip() == '':
        return ()
      if spec.strip() == SmartSort.MARK:
        return SmartSort
      spec = spec.split(',')
    if not morph.isseq(spec):
      raise ValueError(SortValidator.messages['bad_decode'])
    ret = []
    for val in spec:
      if morph.isstr(val):
        val = val.strip()
        ret.append(
          (val[:-1].strip(), False) if val.endswith('-') else (val, True))
        continue
      if len(val) == 2 \
          and morph.isstr(val[0]) \
          and val[1] in (True, False):
        ret.append(val)
        continue
      raise ValueError(SortValidator.messages['bad_decode'])
    return ret
  @staticmethod
  def encode(sort):
    if sort is SmartSort:
      return SmartSort.MARK
    if len(sort) == 0:
      return ''
    return ','.join([
      spec[0] + ( '' if spec[1] else '-' )
      for spec in sort
      if spec[0]])


#------------------------------------------------------------------------------
class Decoder(object):
  '''
  The `Decoder` translates a request into a set of query parameters
  that can be understood by the Paginator.
  '''

  #----------------------------------------------------------------------------
  def __init__(self, request_param='params', structured=False, *args, **kw):
    super(Decoder, self).__init__(*args, **kw)
    self.structured = structured
    self.params     = request_param

  #----------------------------------------------------------------------------
  def extend(self, *args, **kw):
    params = dict(structured=self.structured, request_param=self.params)
    for arg in args:
      params.update(arg)
    params.update(kw)
    return self.__class__(**params)

  #----------------------------------------------------------------------------
  def decode(self, p8n):
    if not getattr(p8n.paginator, 'schema', None):
      p8n.paginator.schema = self.make_schema(p8n)
    ret = p8n.paginator.schema.to_python(
      dict(getattr(p8n.request, self.params)))
    if p8n.paginator.page_name is not None and self.structured:
      ret = ret.get('page')
    return self.validate_sort(p8n, ret)

  #----------------------------------------------------------------------------
  def validate_sort(self, p8n, result):
    if result['sort'] is SmartSort:
      return result
    for spec in result['sort']:
      if spec[0] not in p8n.paginator.engine.comparers:
        self.invalid_sort(p8n, result, spec[0])
    return result

  #----------------------------------------------------------------------------
  def invalid_sort(self, p8n, result, method):
    val = result['sort']
    msg = SortValidator().message('bad_method', None, method=method)
    err = dict([(self.param_name(p8n, p8n.paginator.sort_name), msg)])
    exc = formencode.api.Invalid(
      '\n'.join([k + ': ' + v for k, v in err.items()]),
      val, None, error_dict=err)
    if p8n.paginator.page_name is None or not self.structured:
      raise exc
    raise formencode.api.Invalid(
      self.param_name(p8n, p8n.paginator.page_name) + ': ' + exc.msg,
      val, None, error_dict=dict([
        (self.param_name(p8n, p8n.paginator.page_name), exc)]))

  #----------------------------------------------------------------------------
  def param_name(self, p8n, param):
    if p8n.paginator.page_name is None or self.structured:
      return param
    return '.'.join([p8n.paginator.page_name, param])

  #----------------------------------------------------------------------------
  def make_schema(self, p8n):
    decoder = self
    class PaginationSchema(formencode.Schema):
      filter_extra_fields = True
      allow_extra_fields = True
      chained_validators = [
        RenameFields(dict((
          (decoder.param_name(p8n, p8n.paginator.offset_name), 'offset'),
          (decoder.param_name(p8n, p8n.paginator.limit_name),  'limit'),
          (decoder.param_name(p8n, p8n.paginator.sort_name),   'sort'),
        )))]
      def __init__(self, *args, **kw):
        super(PaginationSchema, self).__init__(*args, **kw)
        self.add_field(
          decoder.param_name(p8n, p8n.paginator.offset_name),
          validators.Int(
            min        = 0,
            if_missing = p8n.paginator.offset_default,
            if_empty   = p8n.paginator.offset_default))
        self.add_field(
          decoder.param_name(p8n, p8n.paginator.limit_name),
          validators.Int(
            min        = 0,
            if_missing = p8n.paginator.limit_default,
            if_empty   = p8n.paginator.limit_default))
        self.add_field(
          decoder.param_name(p8n, p8n.paginator.sort_name),
          SortValidator(
            if_missing = SortValidator.decode(p8n.paginator.sort_default),
            if_empty   = SortValidator.decode('')))
    schema = PaginationSchema()
    if p8n.paginator.page_name is not None and self.structured:
      subschema = PaginationSchema(if_missing=None)
      class PaginationNamespaceSchema(formencode.Schema):
        filter_extra_fields = True
        allow_extra_fields = True
        chained_validators = [
          RenameFields(dict((
            (decoder.param_name(p8n, p8n.paginator.page_name), 'page'),
          )))]
        def __init__(self, *args, **kw):
          super(PaginationNamespaceSchema, self).__init__(*args, **kw)
          self.add_field(p8n.paginator.page_name, subschema)
        def to_python(self, value):
          # todo: there *must* be a more formencode'ish way of doing this...
          ret = super(PaginationNamespaceSchema, self).to_python(value)
          if ret.get('page', None) is None:
            ret['page'] = subschema.to_python({})
          return ret
      schema = PaginationNamespaceSchema()
    return schema

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------
